"""A series of Sphinx extensions to get easy to maintain, numpydoc style documentation.

This module is also an extension itself which simply sets up all included extensions.
"""
from textwrap import indent
from typing import Any, Dict

from get_version import get_version
from sphinx.application import Sphinx


__version__ = get_version(__file__)

metadata = dict(version=__version__, env_version=1, parallel_read_safe=True)

# Can’t seem to be able to do this in numpydoc style:
# https://github.com/sphinx-doc/sphinx/issues/5887
setup_sig_str = """\
Args:
    app: Sphinx app to set this :term:`sphinx:extension` up for

Returns:
    :ref:`Metadata <sphinx:ext-metadata>` for this extension.
"""


def _setup_sig(fn):
    fn.__doc__ += "\n\n" + indent(setup_sig_str, " " * 4)
    return fn


@_setup_sig
def setup(app: Sphinx) -> Dict[str, Any]:
    """Set up all included extensions!"""
    app.setup_extension("scanpydoc.autosummary_generate_imported")
    app.setup_extension("scanpydoc.definition_list_typed_field")
    app.setup_extension("scanpydoc.elegant_typehints")
    app.setup_extension("scanpydoc.rtd_github_links")
    app.setup_extension("scanpydoc.theme")
    return metadata



import sys
from datetime import datetime
from pathlib import Path

from sphinx.application import Sphinx


# Allow importing scanpydoc itself
HERE = Path(__file__).parent
sys.path.insert(0, str(HERE.parent))
import scanpydoc  # noqa


# Clean build env
for file in HERE.glob("scanpydoc.*.rst"):
    file.unlink()


needs_sphinx = "1.7"  # autosummary bugfix
extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",  # needs to be after napoleon
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "scanpydoc",
]

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    jinja=("https://jinja.palletsprojects.com/en/2.10.x/", None),
    sphinx=("https://www.sphinx-doc.org/en/master/", None),
    sphinx_rtd_theme=("https://sphinx-rtd-theme.readthedocs.io/en/stable/", None),
    # examples
    numpy=("https://numpy.org/doc/stable/", None),
    anndata=("https://anndata.readthedocs.io/en/latest/", None),
    pandas=("https://pandas.pydata.org/pandas-docs/stable/", None),
    scipy=("https://docs.scipy.org/doc/scipy/reference/", None),
)

# general information
project = scanpydoc.__name__
author = "Philipp Angerer"
copyright = f"{datetime.now():%Y}, {author}."
version = release = scanpydoc.__version__

master_doc = "index"
templates_path = ["_templates"]

# Generate .rst stubs for modules using autosummary
autosummary_generate = True
# Don’t add module paths to documented functions’ names
add_module_names = False

html_theme = "scanpydoc"
html_context = dict(
    github_user="theislab", github_repo="scanpydoc", github_version="master"
)

# proj/doc/conf.py/../.. → proj
project_dir = Path(__file__).parent.parent


def setup(app: Sphinx):
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )



from typing import Dict, Optional, Union


def example_func(a: Optional[str], b: Union[str, int, None] = None) -> Dict[str, int]:
    """Example function

    Hover over the parameter and return type annotations to see the long versions.

    Args:
        a: An example parameter
        b: Another, with a default
    Returns:
        An example return value
    """
    pass



import importlib.util
import sys
import typing as t
from tempfile import NamedTemporaryFile
from textwrap import dedent

import pytest
from docutils.nodes import document
from docutils.writers import Writer
from sphinx.application import Sphinx
from sphinx.io import read_doc
from sphinx.testing.fixtures import make_app, test_params  # noqa
from sphinx.testing.path import path as STP
from sphinx.util import rst
from sphinx.util.docutils import sphinx_domains


@pytest.fixture
def make_app_setup(make_app, tmp_path) -> t.Callable[..., Sphinx]:
    def make_app_setup(**conf) -> Sphinx:
        (tmp_path / "conf.py").write_text("")
        return make_app(srcdir=STP(tmp_path), confoverrides=conf)

    return make_app_setup


@pytest.fixture
def parse() -> t.Callable[[Sphinx, str], document]:
    def _parse(app: Sphinx, code: str) -> document:
        with NamedTemporaryFile("w+", suffix=".rst", dir=app.env.srcdir) as f:
            f.write(code)
            f.flush()
            app.env.prepare_settings(f.name)
            with sphinx_domains(app.env), rst.default_role(
                f.name, app.config.default_role
            ):
                return read_doc(app, app.env, f.name)

    return _parse


@pytest.fixture
def render() -> t.Callable[[Sphinx, document], str]:
    def _render(app: Sphinx, doc: document) -> str:
        # Doesn’t work as desc is an Admonition and the HTML writer doesn’t handle it
        # print(app.builder.render_partial(doc[1]))
        app.builder.prepare_writing({doc["source"]})
        writer = app.builder.docwriter  # type: Writer
        writer.document = doc
        writer.document.settings = app.builder.docsettings
        writer.translate()
        return writer.output

    return _render


@pytest.fixture
def make_module():
    added_modules = []

    def make_module(name, code):
        assert name not in sys.modules
        spec = importlib.util.spec_from_loader(name, loader=None)
        mod = sys.modules[name] = importlib.util.module_from_spec(spec)
        exec(dedent(code), mod.__dict__)
        added_modules.append(name)
        return mod

    yield make_module

    for name in added_modules:
        del sys.modules[name]



"""Format typehints elegantly and and fix automatically created links.

The Sphinx extension :mod:`sphinx_autodoc_typehints` adds type annotations to functions.
This extension modifies the created type annotations in four ways:

#. It formats the annotations more simply and in line with e.g. :mod:`numpy`.
#. It defines a configuration value ``qualname_overrides`` for ``conf.py``
   that overrides automatically created links. It is used like this::

       qualname_overrides = {
           "pandas.core.frame.DataFrame": "pandas.DataFrame",
           ...,
       }

   The defaults include :class:`anndata.AnnData`, :class:`pandas.DataFrame`,
   :class:`scipy.sparse.spmatrix` and other classes in :mod:`scipy.sparse`.

   It is necessary since :attr:`~definition.__qualname__` does not necessarily match
   the documented location of the function/class.

   Once either `sphinx issue 4826`_ or `sphinx-autodoc-typehints issue 38`_ are fixed,
   this part of the functionality will no longer be necessary.
#. The config value ``annotate_defaults`` (default: :data:`True`) controls if rST code
   like ``(default: `42`)`` is added after the type.
   It sets sphinx-autodoc-typehints’s option ``typehints_defaults`` to ``'braces'``
#. Type annotations for :class:`tuple` return types are added::

       def x() -> Tuple[int, float]:
           \"""
           Returns:
               a: An integer
               b: A floating point number
           \"""

   will render as:

       :Returns: a : :class:`int`
                     An integer
                 b : :class:`float`
                     A floating point number


.. _sphinx issue 4826: https://github.com/sphinx-doc/sphinx/issues/4826
.. _sphinx-autodoc-typehints issue 38: https://github.com/agronholm/sphinx-autodoc-typehints/issues/38

"""
from collections import ChainMap
from functools import partial
from pathlib import Path
from typing import Any, Dict

import sphinx_autodoc_typehints
from docutils.parsers.rst import roles
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.ext.autodoc import ClassDocumenter

from .. import _setup_sig, metadata
from .example import example_func


HERE = Path(__file__).parent.resolve()

qualname_overrides_default = {
    "anndata.base.AnnData": "anndata.AnnData",
    "anndata.core.anndata.AnnData": "anndata.AnnData",
    "anndata._core.anndata.AnnData": "anndata.AnnData",
    "matplotlib.axes._axes.Axes": "matplotlib.axes.Axes",
    "pandas.core.frame.DataFrame": "pandas.DataFrame",
    "pandas.core.indexes.base.Index": "pandas.Index",
    "scipy.sparse.base.spmatrix": "scipy.sparse.spmatrix",
    "scipy.sparse.csr.csr_matrix": "scipy.sparse.csr_matrix",
    "scipy.sparse.csc.csc_matrix": "scipy.sparse.csc_matrix",
}
qualname_overrides = ChainMap({}, qualname_overrides_default)


def _init_vars(app: Sphinx, config: Config):
    qualname_overrides.update(config.qualname_overrides)
    if "sphinx_autodoc_typehints" not in config.extensions and config.annotate_defaults:
        raise ValueError(
            "Can only use annotate_defaults option when using sphinx-autodoc-typehints"
        )
    if config.typehints_defaults is None and config.annotate_defaults:
        # override default for “typehints_defaults”
        config.typehints_defaults = "braces"
    config.html_static_path.append(str(HERE / "static"))


example_func.__module__ = "scanpydoc.elegant_typehints"  # Make it show up here


@_setup_sig
def setup(app: Sphinx) -> Dict[str, Any]:
    """Patches :mod:`sphinx_autodoc_typehints` for a more elegant display."""
    # TODO: Unsure if “html” is sufficient or if we need to do “env”;
    #       Depends on when the autodoc-process-docstring event is handled.
    app.add_config_value("qualname_overrides", {}, "html")
    app.add_config_value("annotate_defaults", True, "html")
    app.add_css_file("typehints.css")
    app.connect("config-inited", _init_vars)

    from .formatting import _role_annot, format_annotation

    sphinx_autodoc_typehints.format_annotation = format_annotation
    for name in ["annotation-terse", "annotation-full"]:
        roles.register_canonical_role(
            name, partial(_role_annot, additional_classes=name.split("-"))
        )

    from .autodoc_patch import dir_head_adder

    ClassDocumenter.add_directive_header = dir_head_adder(
        qualname_overrides, ClassDocumenter.add_directive_header
    )

    from .return_tuple import process_docstring  # , process_signature

    app.connect("autodoc-process-docstring", process_docstring)
    # app.connect("autodoc-process-signature", process_signature)

    return metadata



"""GitHub URLs for class and method pages

This extension registers a :ref:`Jinja filter <jinja:filters>` called :func:`github_url`
that you can use to convert a module path into a GitHub URL

Configuration
-------------

Uses the following config values in ``conf.py``::

    project_dir: Path = ...  # default: Path.cwd()
    html_context = dict(
        github_user=...,
        github_repo=...,
        github_version=...,
    )

The ``project_dir`` is used to figure out the .py file path relative to the git root,
that is to construct the path in the github URL.

The ``html_context`` is e.g. also used like this in the sphinx_rtd_theme_.

Usage
-----

You can use the filter e.g. in `autosummary templates`_.
To configure the sphinx_rtd_theme_, override the ``autosummary/base.rst`` template like this:

.. code:: restructuredtext

    :github_url: {{ fullname | github_url }}

    {% extends "!autosummary/base.rst" %}

.. _autosummary templates: http://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html#customizing-templates
.. _sphinx_rtd_theme: https://sphinx-rtd-theme.readthedocs.io/en/latest/
"""
import inspect
import sys
from pathlib import Path, PurePosixPath
from types import ModuleType
from typing import Any, Dict, Tuple

from jinja2.defaults import DEFAULT_FILTERS
from sphinx.application import Sphinx
from sphinx.config import Config

from . import _setup_sig, metadata


project_dir = None  # type: Path
github_base_url = None  # type: str


def _init_vars(app: Sphinx, config: Config):
    """Called when ``conf.py`` has been loaded"""
    global github_base_url, project_dir
    _check_html_context(config)
    github_base_url = "https://github.com/{github_user}/{github_repo}/tree/{github_version}".format_map(
        config.html_context
    )
    project_dir = Path(config.project_dir)


def _get_obj_module(qualname: str) -> Tuple[Any, ModuleType]:
    """Get a module/class/attribute and its original module by qualname"""
    modname = qualname
    attr_path = []
    while modname not in sys.modules:
        modname, leaf = modname.rsplit(".", 1)
        attr_path.insert(0, leaf)

    # retrieve object and find original module name
    mod = sys.modules[modname]
    obj = None
    for attr_name in attr_path:
        thing = getattr(mod if obj is None else obj, attr_name)
        if isinstance(thing, ModuleType):
            mod = thing
        else:
            obj = thing
            mod_orig = getattr(obj, "__module__", None)
            if mod_orig is not None:
                mod = sys.modules[mod_orig]

    return obj, mod


def _get_linenos(obj):
    """Get an object’s line numbers."""
    try:
        lines, start = inspect.getsourcelines(obj)
    except TypeError:
        return None, None
    else:
        return start, start + len(lines) - 1


def github_url(qualname: str) -> str:
    """Get the full GitHub URL for some object’s qualname.

    Args:
        qualname: The full qualified name of a function, class, method or module

    Returns:
        A GitHub URL derived from the :confval:`html_context`.
    """
    try:
        obj, module = _get_obj_module(qualname)
    except Exception:
        print(f"Error in github_url({qualname!r}):", file=sys.stderr)
        raise
    try:
        path = PurePosixPath(Path(module.__file__).resolve().relative_to(project_dir))
    except ValueError:
        # trying to document something from another package
        path = "/".join(module.__file__.split("/")[-2:])
    start, end = _get_linenos(obj)
    fragment = f"#L{start}-L{end}" if start and end else ""
    return f"{github_base_url}/{path}{fragment}"


def _check_html_context(config: Config):
    try:
        html_context = config.html_context
    except AttributeError:
        raise ValueError(
            f"Extension {__name__} needs “html_context” to be defined in conf.py"
        )
    missing_values = {
        "github_user",
        "github_repo",
        "github_version",
    } - html_context.keys()
    if missing_values:
        mvs = ", ".join([f"html_context[{mv!r}]" for mv in missing_values])
        raise ValueError(
            f"Extension {__name__} needs “{mvs}” to be defined in conf.py.\n"
            f"html_context = {html_context!r}"
        )


@_setup_sig
def setup(app: Sphinx) -> Dict[str, Any]:
    """Register the :func:`github_url` :ref:`Jinja filter <jinja:filters>`."""
    # Guess default project dir
    proj_dir = Path.cwd()
    if proj_dir.name == "docs":
        proj_dir = proj_dir.parent
    elif not (proj_dir / "docs").is_dir():
        proj_dir = proj_dir.parent

    app.add_config_value("project_dir", proj_dir, "")
    app.connect("config-inited", _init_vars)

    # html_context doesn’t apply to autosummary templates ☹
    # and there’s no way to insert filters into those templates
    # so we have to modify the default filters
    DEFAULT_FILTERS["github_url"] = github_url

    return metadata



from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch

from scanpydoc.rtd_github_links import _get_linenos, _get_obj_module, github_url


HERE = Path(__file__).parent


@pytest.fixture
def env(monkeypatch: MonkeyPatch):
    monkeypatch.setattr("scanpydoc.rtd_github_links.github_base_url", ".")
    monkeypatch.setattr("scanpydoc.rtd_github_links.project_dir", HERE)


def test_as_function(env):
    assert github_url("scanpydoc.rtd_github_links") == "./scanpydoc/rtd_github_links.py"
    s, e = _get_linenos(github_url)
    assert (
        github_url("scanpydoc.rtd_github_links.github_url")
        == f"./scanpydoc/rtd_github_links.py#L{s}-L{e}"
    )


def test_get_obj_module():
    import get_version

    obj, mod = _get_obj_module("scanpydoc.get_version")
    assert obj is get_version.get_version
    assert mod is get_version



import pkgutil
from functools import partial
from importlib import import_module

import scanpydoc


def test_all_get_installed(monkeypatch, make_app_setup):
    setups_seen = set()
    setups_called = {}
    for finder, mod_name, _ in pkgutil.walk_packages(
        scanpydoc.__path__, f"{scanpydoc.__name__}."
    ):
        mod = import_module(mod_name)
        if not hasattr(mod, "setup"):
            continue
        setups_seen.add(mod_name)
        monkeypatch.setattr(mod, "setup", partial(setups_called.__setitem__, mod_name))

    app = make_app_setup()
    app.setup_extension("scanpydoc")

    assert setups_called.keys() == setups_seen
    for mod_name, app2 in setups_called.items():
        assert app is app2



import collections.abc as cabc
import inspect
from functools import partial
from typing import Any, Dict, Iterable, List, Sequence, Tuple, Type, Union


try:  # 3.8 additions
    from typing import Literal, get_args, get_origin
except ImportError:
    from typing_extensions import Literal, get_args, get_origin

import sphinx_autodoc_typehints
from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst.roles import set_classes
from docutils.parsers.rst.states import Inliner, Struct
from docutils.utils import SystemMessage, unescape
from sphinx_autodoc_typehints import format_annotation as _format_orig

from scanpydoc import elegant_typehints


def _format_full(
    annotation: Type[Any],
    fully_qualified: bool = False,
    simplify_optional_unions: bool = True,
):
    if inspect.isclass(annotation) and annotation.__module__ == "builtins":
        return _format_orig(annotation, fully_qualified, simplify_optional_unions)

    origin = get_origin(annotation)
    tilde = "" if fully_qualified else "~"

    annotation_cls = annotation if inspect.isclass(annotation) else type(annotation)
    if annotation_cls.__module__ == "typing":
        return _format_orig(annotation, fully_qualified, simplify_optional_unions)

    # Only if this is a real class we override sphinx_autodoc_typehints
    if inspect.isclass(annotation) or inspect.isclass(origin):
        full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        override = elegant_typehints.qualname_overrides.get(full_name)
        role = "exc" if issubclass(annotation_cls, BaseException) else "class"
        if override is not None:
            return f":py:{role}:`{tilde}{override}`"

    return _format_orig(annotation, fully_qualified, simplify_optional_unions)


def _format_terse(
    annotation: Type[Any],
    fully_qualified: bool = False,
    simplify_optional_unions: bool = True,
) -> str:
    origin = get_origin(annotation)
    args = get_args(annotation)
    tilde = "" if fully_qualified else "~"
    fmt = partial(
        _format_terse,
        fully_qualified=fully_qualified,
        simplify_optional_unions=simplify_optional_unions,
    )

    # display `Union[A, B]` as `A | B`
    if origin is Union:
        # Never use the `Optional` keyword in the displayed docs.
        # Use `| None` instead, similar to other large numerical packages.
        return " | ".join(map(fmt, args))

    # do not show the arguments of Mapping
    if origin is cabc.Mapping:
        return f":py:class:`{tilde}typing.Mapping`"

    # display dict as {k: v}
    if origin is dict and len(args) == 2:
        k, v = args
        return f"{{{fmt(k)}: {fmt(v)}}}"

    # display Callable[[a1, a2], r] as (a1, a2) -> r
    if origin is cabc.Callable and len(args) == 2:
        params, ret = args
        params = ["…"] if params is Ellipsis else map(fmt, params)
        return f"({', '.join(params)}) → {fmt(ret)}"

    # display Literal as {'a', 'b', ...}
    if origin is Literal:
        return f"{{{', '.join(map(repr, args))}}}"

    return _format_full(annotation, fully_qualified, simplify_optional_unions)


def format_annotation(
    annotation: Type[Any],
    fully_qualified: bool = False,
    simplify_optional_unions: bool = True,
) -> str:
    r"""Generate reStructuredText containing links to the types.

    Unlike :func:`sphinx_autodoc_typehints.format_annotation`,
    it tries to achieve a simpler style as seen in numeric packages like numpy.

    Args:
        annotation: A type or class used as type annotation.
        fully_qualified: If links should be formatted as fully qualified
            (e.g. ``:py:class:`foo.Bar```) or not (e.g. ``:py:class:`~foo.Bar```).
        simplify_optional_unions: If Unions should be minimized if they contain
            3 or more elements one of which is ``None``. (If ``True``, e.g.
            ``Optional[Union[str, int]]`` becomes ``Union[str, int, None]``)

    Returns:
        reStructuredText describing the type
    """
    if sphinx_autodoc_typehints.format_annotation is not format_annotation:
        raise RuntimeError(
            "This function is not guaranteed to work correctly without overriding"
            "`sphinx_autodoc_typehints.format_annotation` with it."
        )

    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    if calframe[1][3] == "process_docstring":
        return format_both(annotation, fully_qualified, simplify_optional_unions)
    else:  # recursive use
        return _format_full(annotation, fully_qualified, simplify_optional_unions)


def format_both(
    annotation: Type[Any],
    fully_qualified: bool = False,
    simplify_optional_unions: bool = True,
) -> str:
    terse = _format_terse(annotation, fully_qualified, simplify_optional_unions)
    full = _format_full(annotation, fully_qualified, simplify_optional_unions)
    return f":annotation-terse:`{_escape(terse)}`\\ :annotation-full:`{_escape(full)}`"


def _role_annot(
    name: str,
    rawtext: str,
    text: str,
    lineno: int,
    inliner: Inliner,
    options: Dict[str, Any] = {},
    content: Sequence[str] = (),
    # *,  # https://github.com/ambv/black/issues/613
    additional_classes: Iterable[str] = (),
) -> Tuple[List[Node], List[SystemMessage]]:
    options = options.copy()
    set_classes(options)
    if additional_classes:
        options["classes"] = options.get("classes", []).copy()
        options["classes"].extend(additional_classes)
    memo = Struct(
        document=inliner.document, reporter=inliner.reporter, language=inliner.language
    )
    node = nodes.inline(unescape(rawtext), "", **options)
    children, messages = inliner.parse(_unescape(text), lineno, memo, node)
    node.extend(children)
    return [node], messages


def _escape(rst: str) -> str:
    return rst.replace("`", "\\`")


def _unescape(rst: str) -> str:
    # TODO: IDK why the [ part is necessary.
    return unescape(rst).replace("\\`", "`").replace("[", "\\[")



from functools import wraps
from typing import Callable, Mapping, Tuple

from docutils.statemachine import StringList
from sphinx.ext.autodoc import ClassDocumenter


def dir_head_adder(
    qualname_overrides: Mapping[str, str],
    orig: Callable[[ClassDocumenter, str], None],
):
    @wraps(orig)
    def add_directive_header(self: ClassDocumenter, sig: str) -> None:
        orig(self, sig)
        lines: StringList = self.directive.result
        role, direc = (
            ("exc", "exception")
            if isinstance(self.object, type) and issubclass(self.object, BaseException)
            else ("class", "class")
        )
        for old, new in qualname_overrides.items():
            # Currently, autodoc doesn’t link to bases using :exc:
            lines.replace(f":class:`{old}`", f":{role}:`{new}`")
            # But maybe in the future it will
            lines.replace(f":{role}:`{old}`", f":{role}:`{new}`")
            old_mod, old_cls = old.rsplit(".", 1)
            new_mod, new_cls = new.rsplit(".", 1)
            replace_multi_suffix(
                lines,
                (f".. py:{direc}:: {old_cls}", f"   :module: {old_mod}"),
                (f".. py:{direc}:: {new_cls}", f"   :module: {new_mod}"),
            )

    return add_directive_header


def replace_multi_suffix(lines: StringList, old: Tuple[str, str], new: Tuple[str, str]):
    if len(old) != len(new) != 2:
        raise NotImplementedError("Only supports replacing 2 lines")
    for l, line in enumerate(lines):
        start = line.find(old[0])
        if start == -1:
            continue
        prefix = line[:start]
        suffix = line[start + len(old[0]) :]
        if lines[l + 1].startswith(prefix + old[1]):
            break
    else:
        return
    lines[l + 0] = prefix + new[0] + suffix
    lines[l + 1] = prefix + new[1]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['scanpydoc', 'scanpydoc.elegant_typehints', 'scanpydoc.theme']

package_data = \
{'': ['*'],
 'scanpydoc.elegant_typehints': ['static/*'],
 'scanpydoc.theme': ['static/*', 'static/css/*']}

install_requires = \
['sphinx>=3.0', 'get_version']

extras_require = \
{'dev': ['pre-commit'],
 'doc': ['scanpydoc[typehints]', 'sphinx-rtd-theme'],
 'test': ['pytest', 'pytest-cov'],
 'test:python_version<"3.8"': ['typing_extensions>=3.10'],
 'theme': ['sphinx-rtd-theme'],
 'typehints': ['sphinx-autodoc-typehints>=1.14']}

entry_points = \
{'sphinx.html_themes': ['scanpydoc = scanpydoc.theme']}

setup(name='scanpydoc',
      version='0.7.4',
      description='A series of Sphinx extensions to get easy to maintain, numpydoc style documentation.',
      author='Philipp Angerer',
      author_email='phil.angerer@gmail.com',
      url='https://github.com/theislab/scanpydoc/',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



import inspect
import re
from logging import getLogger
from typing import Any, List, Optional, Tuple, Type, Union, get_type_hints


try:  # 3.8 additions
    from typing import get_args, get_origin
except ImportError:
    from typing_extensions import get_args, get_origin

from sphinx.application import Sphinx
from sphinx.ext.autodoc import Options

from .formatting import format_both


logger = getLogger(__name__)
re_ret = re.compile("^:returns?: ")


def get_tuple_annot(annotation: Optional[Type]) -> Optional[Tuple[Type, ...]]:
    if annotation is None:
        return None
    origin = get_origin(annotation)
    if not origin:
        return None
    if origin is Union:
        for annotation in get_args(annotation):
            origin = get_origin(annotation)
            if origin in (tuple, Tuple):
                break
        else:
            return None
    return get_args(annotation)


def process_docstring(
    app: Sphinx,
    what: str,
    name: str,
    obj: Any,
    options: Optional[Options],
    lines: List[str],
) -> None:
    # Handle complex objects
    if isinstance(obj, property):
        obj = obj.fget
    if not callable(obj):
        return
    if what in ("class", "exception"):
        obj = obj.__init__
    obj = inspect.unwrap(obj)
    try:
        hints = get_type_hints(obj)
    except (AttributeError, TypeError):
        # Introspecting a slot wrapper will raise TypeError
        return
    except NameError as e:
        check_bpo_34776(obj, e)
        return
    ret_types = get_tuple_annot(hints.get("return"))
    if ret_types is None:
        return

    # Get return section
    i_prefix = None
    l_start = None
    l_end = None
    for l, line in enumerate(lines):
        if i_prefix is None:
            m = re_ret.match(line)
            if m:
                i_prefix = m.span()[1]
                l_start = l
        elif len(line[:i_prefix].strip()) > 0:
            l_end = l - 1
            break
    else:
        l_end = len(lines) - 1
    if i_prefix is None:
        return

    # Meat
    idxs_ret_names = []
    for l, line in enumerate([l[i_prefix:] for l in lines[l_start : l_end + 1]]):
        if line.isidentifier() and lines[l + l_start + 1].startswith("    "):
            idxs_ret_names.append(l + l_start)

    if len(idxs_ret_names) == len(ret_types):
        for l, rt in zip(idxs_ret_names, ret_types):
            typ = format_both(rt, app.config.typehints_fully_qualified)
            lines[l : l + 1] = [f"{lines[l]} : {typ}"]


# def process_signature(
#     app: Sphinx,
#     what: str,
#     name: str,
#     obj: Any,
#     options: Options,
#     signature: Optional[str],
#     return_annotation: str,
# ) -> Optional[Tuple[Optional[str], Optional[str]]]:
#     return signature, return_annotation


def check_bpo_34776(obj: Any, e: NameError):
    import sys

    ancient = sys.version_info < (3, 7)
    old_3_7 = (3, 7) < sys.version_info < (3, 7, 6)
    old_3_8 = (3, 8) < sys.version_info < (3, 8, 1)
    if ancient or old_3_7 or old_3_8:
        v = ".".join(map(str, sys.version_info[:3]))
        logger.warning(
            f"Error documenting {obj!r}: To avoid this, "
            f"your Python version {v} must be at least 3.7.6 or 3.8.1. "
            "For more information see https://bugs.python.org/issue34776"
        )
    else:
        raise e  # No idea what happened
